// file: dated.h
// author: Bernt A Oedegaard.
// 
// dated<T> , Dated Elements, a list of dates with accompanying Elements of any type<T>

#ifndef _DATED_H_
#define _DATED_H_

#include <vector>     // vector templates
#include <algorithm>
#include "date.h"       // my date class

template <class T> class dated {
private:
    vector<date> dates_;
    vector<T> elements_;
public:
    dated<T>();
    dated<T>(const dated<T>&);
    dated<T> operator= (const dated<T>&);
    ~dated() { clear(); }; 
    void append(const date&, const T&); // adding at the back.
    void insert(const date&, const T&); // insert somewhere

    void clear();                     // erasing
    void remove(const date&);         // removing one or more elements
    void remove_between(const date&, const date&);
    void remove_before(const date&);
    void remove_after(const date&);
public:  // when searching in the data, these are useful functions
    int index_of_date(const date& d) const;
    int index_of_last_date_before(const date& d) const;
    int index_of_first_date_after(const date& d) const;
public:            // pulling data
    bool empty() const ;
    int  size() const ;
    bool contains(const date& d) const ;
    date  date_at(const int& t) const ;
    T element_at(const int& t) const ;
    T element_at(const date& d) const ;

    // next: the element either on date d, if d is here, else the last observation before d.
    T current_element_at(const date& d) const ;

    vector<T> elements() const; // vectors of <>

    vector<date> dates() const { 
	vector<date> ds(size());
	for (int t=0;t<size();++t){ ds[t]=date_at(t); };
	return ds;
    };

    date first_date() const;  // simple queries
    date last_date() const;
    T first_element() const;
    T last_element() const;
};

////////////////////////////////////////////
// implementation //////////////////////////
////////////////////////////////////////////

template<class T> bool dated<T>::empty() const { return (dates_.size()<1); };  
template<class T> int  dated<T>::size() const { return dates_.size(); };
template<class T> bool dated<T>::contains(const date& d) const {
   return binary_search(dates_.begin(),dates_.end(),d);
};

template<class T> date dated<T>::date_at(const int& t) const { // accessing with counds checking 
   if ( (t>=0) && (t<size()) ) return dates_[t];
   return date();
};

template<class T> T dated<T>::element_at(const int& t) const {  // accessing with counds checking 
   if ( (t>=0) && (t<size()) ) return elements_[t]; 
   return T();
};

template<class T> T dated<T>::element_at(const date& d) const {
   if (!contains(d)) return T();
   return elements_[index_of_date(d)];
};

// the element either on date d, if d is here, else the last observation before d.
template<class T> T dated<T>::current_element_at(const date& d) const {
    if (size()<1) return T();
    if (contains(d)) return element_at(d);
    if (d<first_date()) { return T(); };
    return elements_[index_of_last_date_before(d)];
};

template<class T>  vector<T> dated<T>::elements() const {
   vector<T> elements(size());
   for (int t=0; t<size(); ++t){ elements[t]=element_at(t); };
   return elements;
};;

template<class T>  date dated<T>::first_date() const { 
   if (empty()) return date(); 
   return dates_.front();
};

template<class T> date dated<T>::last_date() const { 
   if (empty()) return date(); 
   return dates_.back();
};

template<class T> T dated<T>::first_element() const { 
   if (empty()) return T(); 
   return elements_.front();
};

template<class T> T dated<T>::last_element() const { 
   if (empty()) return T(); 
   return elements_.back();
};

template<class T> dated<T>::dated(){
   // not necessary to do anything, 
};

template<class T>dated<T>::dated(const dated<T>& dobs) {
   // for speed, initialize first with correct size and then copy
   dates_=vector<date>(dobs.size());
   elements_=vector<T>(dobs.size());
   for (int t=0;t<dobs.size();++t){
      dates_[t]      = dobs.date_at(t);
      elements_[t]   = dobs.element_at(t);
   };
};

template<class T> dated<T> dated<T>::operator= (const dated<T>& dobs) {
   if (this==&dobs) return *this; // check against self assignment;
   clear();
   dates_=vector<date>(dobs.size());
   elements_=vector<T>(dobs.size());
   for (int t=0;t<dobs.size();++t){
      dates_[t]        = dobs.date_at(t);
      elements_[t]     = dobs.element_at(t);
   };   
   return *this;
};

template <class T> void dated<T>::append(const date& d, const T& obs) {
   // synonym of insert, won't bother about the possible efficiency gains
   // at the cost of possibly getting things out of order.
   // Appending is efficient in the insert code anyway (two tests)
   insert(d,obs);
};

template <class T> void dated<T>::insert(const date& d, const T& obs) {
    if ( (empty() ) || (d>last_date()) ) {
	dates_.push_back(d);
	elements_.push_back(obs);
	return;
    };
    if (d<first_date()) {
	dates_.insert(&dates_[0],d);
	elements_.insert(&elements_[0],obs);
	return;
    };
    if (contains(d)) {
       elements_[index_of_date(d)]=obs;
       return;
    };
    int i = index_of_first_date_after(d);
    dates_.insert(&dates_[i],d);
    elements_.insert(&elements_[i],obs); 
    return;
};

template <class T> int dated<T>::index_of_date(const date& d) const {
   // this routine returns the index at which date d is, or -1 if not found.
   if (!d.valid()) return -1;
   if (!contains(d)) return -1;
   int dist=0;
   distance (dates_.begin(), lower_bound(dates_.begin(),dates_.end(),d), dist);
   return dist;
};

template <class T> int dated<T>::index_of_first_date_after(const date& d) const {
   // this routine returns the index of the first date after d.
   if (d>=last_date()) return -1;
   if (d<first_date()) return 0;
   int dist=0;
   distance (dates_.begin(), upper_bound(dates_.begin(),dates_.end(),d), dist);
   if (dates_[dist]==d) ++dist;
   return dist;
};

template <class T> int dated<T>::index_of_last_date_before(const date& d) const {
   // this routine returns the index of the first date after d.
   if (d<=first_date()) return -1;
   if (d>last_date()) return index_of_date(last_date());
   int dist=0;
   distance (dates_.begin(),lower_bound(dates_.begin(),dates_.end(),d), dist);
   return --dist;
};

template<class T> void dated<T>::clear() {
   dates_    .erase(dates_    .begin(),dates_    .end());
   elements_ .erase(elements_ .begin(),elements_ .end());
};

template<class T> void dated<T>::remove(const date& d) {
   int i=index_of_date(d);
   if (i>=0) {
      dates_.erase(&dates_[i]);
      elements_.erase(&elements_[i]);
   };
};

template<class T> void dated<T>::remove_between(const date& d1, const date& d2) {
   for (int t=size()-1;t>=0;t--){  // this is very slow, to be replaced with one using vector erase.
      date d=date_at(t);
      if ( (d>d1) && (d<d2) ) { 
	 dates_.erase(&dates_[t]);
	 elements_.erase(&elements_[t]);
      };
   };
};

template<class T> void dated<T>::remove_before(const date& d) {
   remove_between(--first_date(),d);
};

template<class T> void dated<T>::remove_after(const date& d) {
   remove_between(d,++last_date());
};

template <class T> ostream& operator << (ostream& outf, const dated<T>& dobs ) {
   if (dobs.empty()) return outf;
   for (int t=0; t<dobs.size(); ++t) {
      outf << dobs.date_at(t) 
	   << " " 
	   << dobs.element_at(t)
	   << endl;
   };
   return outf;
};


#endif 
