// file: security_price_hist.cc
// author: Bernt A Oedegaard
// implementation of security price history
// Prices are positive.  Negative prices are viewed as missing observations.

#include "security_price_history.h"  
#include <fstream>
#include <algorithm>

const double MISSING_OBS = -1;

/////////// basic operations ///////////////////////////////////

security_price_history::security_price_history(){ clear(); }; // make sure empty
security_price_history::security_price_history(const string& name){
   clear(); 
   security_name_ = name; 
};

// copy construction
security_price_history::security_price_history(const security_price_history& sh) {
   clear();
   dates_  = vector<date>(sh.size());
   bids_   = vector<double>(sh.size());
   trades_ = vector<double>(sh.size());
   asks_   = vector<double>(sh.size());
   for (unsigned i=0;i<sh.no_dates();++i){
      dates_[i]=sh.date_at(i);
      bids_[i]=sh.bid(i);
      trades_[i]=sh.trade(i);
      asks_[i]=sh.ask(i);
    };
    set_company_name(sh.company_name()); 
    set_security_name(sh.security_name()); 
};

security_price_history security_price_history:: operator=  (const security_price_history& sh) {
   clear();
   dates_ = vector<date>(sh.size());
   bids_  = vector<double>(sh.size());
   trades_= vector<double>(sh.size());
   asks_  = vector<double>(sh.size());
   for (unsigned i=0;i<sh.no_dates();++i){
      dates_[i]=sh.date_at(i);
      bids_[i]=sh.bid(i);
      trades_[i]=sh.trade(i);
      asks_[i]=sh.ask(i);
    };
    set_company_name(sh.company_name()); 
    set_security_name(sh.security_name()); 
    return (*this);
};

// desctruction
security_price_history::~security_price_history(){ clear(); }; // make sure empty

void security_price_history::clear_prices() {
    dates_.erase(dates_.begin(),dates_.end());
    bids_.erase(bids_.begin(),bids_.end());
    trades_.erase(trades_.begin(),trades_.end());
    asks_.erase(asks_.begin(),asks_.end());
};

void security_price_history::clear() {
    security_name_=string();
    company_name_=string();
    clear_prices();
};

bool security_price_history::empty() const{
    if ( (no_dates()<1)
	 && (company_name().length()<1) 
	 && (security_name().length()<1) 
       )
       return true;
    return false;
};

bool security_price_history::contains(const date& d) const {
  return binary_search(dates_.begin(),dates_.end(),d);
};

int security_price_history::index_of_date(const date& d) const {
   // this routine returns the index at which date d is, or -1 if not found.
   if (!contains(d)) return -1;
   int dist=0;
   distance (dates_.begin(),
	     lower_bound(dates_.begin(), dates_.end(), d),
	     dist);
   return dist;
};

int security_price_history::index_of_first_date_after(const date& d) const {
   if (!d.valid()) return -1;
   if (d>=last_date()) return -1;
   if (d<first_date()) return 0;
   int dist=0;
   distance (dates_.begin(), upper_bound(dates_.begin(),dates_.end(),d), dist);
   if (dates_[dist]==d) ++dist;
   return dist;
};

int security_price_history::index_of_last_date_before(const date& d) const {
   if (!d.valid()) return -1;

   if (d<=first_date()) return -1;
   if (d>last_date()) return index_of_date(last_date());
   int dist=0;
   distance (dates_.begin(),lower_bound(dates_.begin(),dates_.end(),d), dist);
   return --dist;
};

///// price functionss /////////////////////////////////////////////

void security_price_history::add_prices(const date& d,
					const double& bid,
					const double& trade,
					const double& ask) {
   if (!d.valid()) return;  // don't add 
   if ( (bid<0) && (trade<0) && (ask<0) ) return;  // don't bother about empty dates
   if ( (no_dates()<1) || (d>last_date() ) ) {
      dates_ .push_back(d);
      trades_.push_back(trade);
      bids_  .push_back(bid);
      asks_  .push_back(ask);
      return;
   };
   if  (d<first_date()) {
      dates_.insert(dates_.begin(),d);
      bids_.insert(bids_.begin(),bid); 
      trades_.insert(trades_.begin(),trade);
      asks_.insert(asks_.begin(),ask);
   };
   int i = index_of_date(d);
   if (i>=0) { // found, replace
      trades_[i]=trade;
      bids_[i]=bid;
      asks_[i]=ask;
      return;
   };
   // evidently should be inserted somewhere in between other observations.
   for (i=0;i<no_dates();++i){
      if (date_at(i)>d) {
	 dates_.insert(&dates_[i],d);
	 bids_.insert(&bids_[i],bid); 
	 trades_.insert(&trades_[i],trade);
	 asks_.insert(&asks_[i],ask);
	 return;
      };
   };
};

void security_price_history::erase(const date& d) {
   int i = index_of_date(d);
   cout << "erasing" << endl;
   if (i>=0) { // returns -1 if not there
      cout << "erasing" << endl;
      dates_.erase(&dates_[i]);
      trades_.erase(&trades_[i]);
      bids_.erase(&bids_[i]); 
      asks_.erase(&asks_[i]);
   };
};

void security_price_history::erase_between(const date& first,
					   const date& last) {
   int i1 = index_of_first_date_after(first);
   int i2 = index_of_last_date_before(last);
   if ( (i1>=0) && (i2>=0) ) { // returns -1 if not there
      i2++;  // the erase does not delete the last one.
      dates_.erase(&dates_[i1],&dates_[i2]);
      trades_.erase(&trades_[i1],&trades_[i2]);
      bids_.erase(&bids_[i1],&bids_[i2]);
      asks_.erase(&asks_[i1],&asks_[i2]);
   };
};

void security_price_history::erase_before(const date& d) {
   int i = index_of_last_date_before(d);
   if (i>=0)  { // returns -1 if not there
      i++;// the erase does not delete the last one, so add one.
      dates_.erase(dates_.begin(),&dates_[i]);
      trades_.erase(trades_.begin(),&trades_[i]);
      bids_.erase(bids_.begin(),&bids_[i]);
      asks_.erase(asks_.begin(),&asks_[i]);
   };
};

void security_price_history::erase_after(const date& d) {
   int i = index_of_first_date_after(d);
   if (i>=0)  { // returns -1 if not there
      dates_.erase(&dates_[i],dates_.end());
      trades_.erase(&trades_[i],trades_.end());
      bids_.erase(&bids_[i],bids_.end());
      asks_.erase(&asks_[i],asks_.end());
   };
};

double security_price_history::price(const unsigned& i) const{
    if (i>=no_dates()) { return MISSING_OBS; };
    if (trade(i)>0.0) { return trade(i); };
    if ( (bid(i)>0.0)&&(ask(i)>0.0))     // return bid ask average
	{ return 0.5*(bid(i)+ask(i)); };
//    if (bid(i)>0.0)   { return bid(i); };  // to avoid big jumps, 
//    if (ask(i)>0.0)   { return ask(i); };  // skip these
    return MISSING_OBS;
};

double security_price_history::bid(const date& d) const {
   // return price on given date, only if we have an observation on that date.
   // if not, return missing
  int i = index_of_date(d); 
  if (i>=0) return bid(i);
  return MISSING_OBS;
};

double security_price_history::trade(const date& d) const {
   // return price on given date, only if we have an observation on that date.
   // if not, return missing
  int i = index_of_date(d); 
  if (i>=0) return trade(i);
  return MISSING_OBS;
};

double security_price_history::ask(const date& d) const {
   // return price on given date, only if we have an observation on that date.
   // if not, return missing
  int i = index_of_date(d); 
  if (i>=0) return ask(i);
  return MISSING_OBS;
};

double security_price_history::price(const date& d) const {
   // return price on given date, only if we have an observation on that date.
   // if not, return missing
  int i = index_of_date(d); 
  if (i>=0) return price(i);
  return MISSING_OBS;
};

double security_price_history::current_price(const date& d) const {
   // return price on given date.  
   // If don't have that price, return the last one observed
   //  before the wanted date
   // if before first or after last, return missing
   if (empty())         { return MISSING_OBS; };
   if (d<first_date())  { return MISSING_OBS; };
   if (d>last_date())   { return MISSING_OBS; };

   int i = index_of_date(d);  
   if (i>=0) return price(i);

   int ind=0;    // otherwise use last previously observed price.
   distance(dates_.begin(),upper_bound(dates_.begin(),dates_.end(),d),ind);
   return price(--ind);
};

double security_price_history::BuyPrice(const unsigned& i) const{
  if (i>=no_dates()) { return MISSING_OBS; };
  if (ask(i)>0.0)   { return ask(i); };
  if (trade(i)>0.0) { return trade(i); };
  return MISSING_OBS;
};

double security_price_history::SellPrice(const unsigned& i) const{
  if (i>=no_dates()) { return MISSING_OBS; };
  if (bid(i)>0.0)   { return bid(i); };
  if (trade(i)>0.0) { return trade(i); };
  return MISSING_OBS;
};


//////////////////////////////////////////////////////////////////////////
//// summarizing no of prices ////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

unsigned security_price_history::no_prices() const {
  int n_prices=0;
  for (unsigned i=0;i<no_dates();++i){
    if ( (bid(i)>0) || (trade(i)>0) || (ask(i)>0) ) { ++n_prices; };
  };
  return n_prices;
};

unsigned security_price_history::no_prices_between(const date& first,
						   const date& last) const 
{
  unsigned n_pric_between=0;
  for (unsigned i=0;i<no_dates();++i) 
    { 
      if  ((date_at(i)>=first) && (date_at(i)<=last)) 
	{
	  if((bid(i)>0)||(trade(i)>0)||(ask(i)>0)) {++n_pric_between;};
	};
    };
  return n_pric_between;
};

unsigned security_price_history::no_bids() const {
  unsigned n_prices=0;
  for (unsigned i=0;i<no_dates();++i){
    if (bid(i)>0) { ++n_prices; };
  };
  return n_prices;
};


unsigned security_price_history::no_trades() const {
  unsigned n_prices=0;
  for (unsigned i=0;i<no_dates();++i){
    if (trade(i)>0) { ++n_prices; };
  };
  return n_prices;
};

unsigned security_price_history::no_asks() const  {
   unsigned n_prices=0;
   for (unsigned i=0;i<no_dates();++i){
      if (ask(i)>0) { ++n_prices; };
   };
   return n_prices;
};

/// extraction
string security_price_history::company_name()     const { return company_name_;    };
string security_price_history::security_name()    const { return security_name_;    };

unsigned security_price_history::no_dates ()  const { return dates_.size(); };

date security_price_history::first_date() const { 
    if (dates_.size()>0) { return dates_.front();}; 
    return date(); // else
};

 date security_price_history::last_date() const { 
    if (dates_.size()>0) { return dates_.back();};
    return date();
};

////////////////////// changing and adding data /////////////////////

void security_price_history::set_company_name(const string& s)     { company_name_ = s; };
void security_price_history::set_security_name(const string& s)    { security_name_ = s; };

void security_price_history::set_bid(unsigned i, double bid)     { if (i<no_dates()) bids_[i]=bid; };
void security_price_history::set_trade(unsigned i, double trade) { if (i<no_dates()) trades_[i]=trade; };
void security_price_history::set_ask(unsigned i, double ask)     { if (i<no_dates()) asks_[i]=ask; };


/////////////////////////////////////////////////////////////////////////
/// non member functions ////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////

dated_observations prices(const security_price_history& sh ){ 
    dated_observations pr;
    for (int i=0;i<sh.no_prices();++i) {
	date d=sh.date_at(i);
	double price = sh.price(i);
	if (price>0) pr.append(d,price);
    };
    return pr;
};

dated_observations returns(const security_price_history& sh ){ 
    dated_observations pr = prices(sh);
    return returns_given_prices(pr);
};

void dump(ostream& of, const security_price_history& sh ){ 
   if (sh.security_name().length()>0)  of << " Security: " << sh.security_name() << endl;
   print_prices(of,sh);
};

void print_prices(ostream& of, const security_price_history& sh) {
   if (sh.no_dates()>0) {
      of << " date    bid  trade  ask " << endl;
      for (unsigned i=0; i<sh.no_dates(); ++i) { 
	 of << " " ;
	 //	    of << i << " " ;
	 of << sh.date_at(i)  << " ";
	 if (sh.bid(i)>0) { of << sh.bid(i); } else { of << " * ";}; 
	 of << " ";
	 if (sh.trade(i)>0) { of << sh.trade(i); } else { of << " * ";};
	 of << " ";
	 if (sh.ask(i)>0) { of << sh.ask(i); } else { of << " * ";}; 
	 of << endl;
      };
   }
   else {
      of << " no price observations " << endl;
   };
   of << "**********************" << endl;
};



