// file security_price_history.h
// author: Bernt Arne Oedegaard
// define a class that stores a price history of a security. 

#ifndef _SECURITY_PRICE_HISTORY_H_
#define _SECURITY_PRICE_HISTORY_H_

#include <vector>   // vector templates 
#include <string>    // standard string library
#include "date.h"
#include "dated_obs.h"

class security_price_history {
   private:
      string         company_name_;
      string         security_name_;
      vector<date>   dates_;
      vector<double> bids_;
      vector<double> trades_;
      vector<double> asks_;
   public: 
      security_price_history();
      security_price_history(const string&);
      security_price_history(const security_price_history&); 
      security_price_history operator=(const security_price_history&);
      ~security_price_history();
      void clear();
      void clear_prices();

      void set_company_name(const string&);
      void set_security_name(const string&);

      // the most important function, all dates is added through calls to add_prices
      void add_prices(const date& d, 
		      const double& bid, const double& trade, const double& ask);

      void set_bid   (unsigned i, double bid);  // to change current data.
      void set_trade (unsigned i, double trade);
      void set_ask   (unsigned i, double ask);

      void erase(const date&);  // delete on given dates
      void erase_before(const date&);
      void erase_between(const date&, const date&);
      void erase_after(const date&);
      bool contains(const date&) const;  // check whether this date is present
   private:
      int index_of_date(const date&) const;  /* searching through dates and
					      * return where in vector it is 
					      */
      int index_of_last_date_before(const date&) const ;
      int index_of_first_date_after(const date&) const ;
   public:
      bool empty() const;
      int size() const { return dates_.size(); };

      string company_name() const;
      string security_name() const; 

      unsigned no_dates() const;
      unsigned no_prices() const;
      unsigned no_bids() const;
      unsigned no_trades() const;
      unsigned no_asks() const;
      unsigned no_prices_between(const date&, const date&) const;

      date date_at(const unsigned) const;  
      double bid(const unsigned) const; 
      double bid(const date& ) const;
      double trade(const unsigned) const;
      double trade(const date&) const;
      double ask(const unsigned) const;
      double ask(const date&) const;

      date first_date() const;
      date last_date() const;
      double price(const date&) const;  // price AT given data
      double current_price(const date&) const; // price AT or BEFORE given data

      double price(const unsigned&) const; 

      double BuyPrice(const unsigned&) const;
      double SellPrice(const unsigned&) const; 
};

/// non member functions //////////////////////////

void dump                   (ostream&, const security_price_history&);
void print_prices           (ostream&, const security_price_history&);

dated_observations prices(const security_price_history&);
dated_observations returns(const security_price_history&);
dated_observations wealth_relatives(const security_price_history&);

///////////// inline functions //////////////////////////////////////////

//////////// prices ////////////////////////////////////////////////////////
/// note: no bounds checking, the user is assumed to know vector sizes /////

inline date   security_price_history:: date_at(unsigned t)  const { return dates_[t]; };
inline double security_price_history:: bid (unsigned t)  const { return bids_[t]; };
inline double security_price_history:: trade(unsigned t) const { return trades_[t]; };
inline double security_price_history:: ask (unsigned t)  const { return asks_[t]; };

#endif
