// file stock_price_hist.cc
// author: Bernt A Oedegaard
// implementation of stock calculations.
// Some basic assumptions: 
// Prices are positive.  Negative prices are viewed as missing observations.
// this file contains code for
//     constructors
//     destructors
//     basic operations
//     price operations
//     dividend operations
//     adjustment operations


#include "stock_price_history.h"  

const double MISSING_OBS = -1;

/////////// basic operations ///////////////////////////////////

stock_price_history::stock_price_history(){; }; 

stock_price_history::stock_price_history(const stock_price_history& sh):
   security_price_history(sh)
{
   dividends_.clear();
   for (unsigned i=0;i<sh.no_dividends();++i)
      add_dividend(sh.dividend_date(i), sh.dividend(i));
   adjustments_.clear();
   for (unsigned i=0;i<sh.no_adjustments();++i)  
      add_adjustment(sh.adjustment_date(i), sh.adjustment(i));
};

stock_price_history stock_price_history:: operator= (const stock_price_history& sh) {
   clear();
   security_price_history::operator=(sh);
   for (unsigned i=0;i<sh.no_dividends();++i)
      add_dividend(sh.dividend_date(i), sh.dividend(i));
   for (unsigned i=0;i<sh.no_adjustments();++i)  
      add_adjustment(sh.adjustment_date(i), sh.adjustment(i));
   return *this;
};

void stock_price_history::clear() {
    security_price_history::clear();
    dividends_.clear();
    adjustments_.clear();
};

bool stock_price_history::empty() const {
   if (no_dividends()>0) return false;
   if (no_adjustments()>0) return false;
   return security_price_history::empty();
};

///// dividend functionss /////////////////////////////////////////////

void stock_price_history::add_dividend(const date&   dividend_date, 
				       const double& dividend_amount) {
   dividends_.insert(dividend_date,dividend_amount);
};

void stock_price_history::remove_dividend(const date&   dividend_date) { 
   dividends_.remove(dividend_date);
};

bool stock_price_history::dividends_between(const date& d1, const date& d2) const {
    date first, last;
    if (d1<d2) { first=d1; last=d2; } else { first=d2; last=d1; };
    if( dividends_.no_obs_between(first,last)>0) return true;
    return false;
};

int stock_price_history::no_dividends_between(const date& d1, const date& d2)  const {
    date first, last;
    if (d1<d2) { first=d1; last=d2; } else { first=d2; last=d1; };
    return dividends_.no_obs_between(first,last);
};

double stock_price_history::total_dividends_between(const date& d1, const date& d2) const {
   if (!dividends_between(d1,d2)) return 0;
   date first, last;
   if (d1<d2) { first=d1; last=d2; } else { first=d2; last=d1; };
   double tot_dividend = 0.0;
   for (int i=0; i<dividends_.size(); ++i){
      if ( (dividends_.date_at(i)>first) && (dividends_.date_at(i)<=last) ) {
	 tot_dividend += dividends_.element_at(i);
      };
   };
   return tot_dividend;
};

bool stock_price_history::dividend_on(const date& d) const {
   return dividends_.contains(d);
};

double stock_price_history::dividend(const date& d) const {
   double div = dividends_.element_at(d);
   if (div>0) return div;
   return 0.0;
};

unsigned stock_price_history::no_dividends() const    {
   return dividends_.size();
};

date stock_price_history::dividend_date(int i) const {
   return dividends_.date_at(i);
};

double stock_price_history::dividend(int i) const {
   return dividends_.element_at(i);
};

///// adjustment functions /////////////////////////////////////////////

unsigned int stock_price_history::no_adjustments()  const{ 
   return adjustments_.size();
};

date stock_price_history::adjustment_date(const int& i) const{ 
   return adjustments_.date_at(i);
};

double stock_price_history::adjustment(const int& i)  const { 
   return adjustments_.element_at(i);
};

void stock_price_history::add_adjustment ( const date& adjustment_date,
					   const double& adj_factor) {
   if (adjustments_.contains(adjustment_date)) {
      adjustments_.insert(adjustment_date,
			  adjustments_.element_at(adjustment_date)*adj_factor);
   }
   else {
      adjustments_.insert(adjustment_date,adj_factor);
   };
};

void stock_price_history::remove_adjustment ( const date& adjustment_date){
   adjustments_.remove(adjustment_date);
};


bool stock_price_history::adjustments_between(const date& d1, const date& d2)  const{
    date first, last;
    if (d1<d2) { first=d1; last=d2; } else { first=d2; last=d1; };
    if (adjustments_.no_obs_between(first,last)>0) return true;
    return false;
};

int stock_price_history::no_adjustments_between(const date& d1, const date& d2) const {
    date first, last;
    if (d1<d2) { first=d1; last=d2; } else { first=d2; last=d1; };
    return adjustments_.no_obs_between(first,last);
};

double stock_price_history::aggregated_adjustments_between(const date& d1,
						     const date& d2)  const {
    date first, last;
    if (d1<d2) { first=d1; last=d2; } else { first=d2; last=d1; };
    double tot_adjustments=1.0;
    for (int i=0; i<adjustments_.size(); ++i){
	if ( (adjustments_.date_at(i)>first) && (adjustments_.date_at(i)<=last) ) {
	   tot_adjustments *= adjustments_.element_at(i);
	};
    };
    return tot_adjustments;
};

bool stock_price_history::adjustment_on(const date& d) const {
   for (int i=0; i<adjustments_.size(); ++i) {
      if (adjustments_.date_at(i)==d) return true; 
   };
   return false;
};

double stock_price_history::adjustment(const date& d) const {
    for (int i=0; i<adjustments_.size(); ++i){
       if (adjustments_.date_at(i)==d) return adjustments_.element_at(i);
    };
    return 0.0;
};

void dump(ostream& of, const stock_price_history& st){ 
   of << " Stock: " 
      << st.security_name() 
      << endl;
   of << "# Prices: " << endl;
   for (int t=0; t<st.no_prices(); ++t) { 
      of << st.date_at(t)  << " ";
      of << st.bid(t)   << " ";
      of << st.trade(t) << " ";
      of << st.ask(t)   << " ";
      of << endl; 
   };
   if (st.no_dividends()>0) {
      of << "# Dividends " << endl;
      for (int t=0; t<st.no_dividends(); ++t) { 
	 of << " " << st.dividend_date(t) << " ";
	 of << st.dividend(t) << endl;
      };
   };
   if (st.no_adjustments()>0) {
      of << "# Adjustments " << endl;
      for (int t=0; t<st.no_adjustments(); ++t) { 
	 of << " " << st.adjustment_date(t) << " ";
	 of << st.adjustment(t) << endl;
      };
   };
};
